/***************************************************************************
**
**  This file is part of QGpCoreMath.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This file is distributed in the hope that it will be useful, but WITHOUT
**  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
**  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
**  License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**
**  See http://www.geopsy.org for more information.
**
**  Created: 2002-04-15
**  Copyright: 2002-2019
**    Marc Wathelet
**    Marc Wathelet (ULg, Liège, Belgium)
**    Marc Wathelet (LGIT, Grenoble, France)
**
***************************************************************************/

#ifndef POINT2D_H
#define POINT2D_H

#include <QGpCoreTools.h>

#include "Angle.h"
#include "QGpCoreMathDLLExport.h"
#include "SamplingParameters.h"

namespace QGpCoreMath {

class CurvePointOptions;
class Matrix2x2;
class UtmZone;

class QGPCOREMATH_EXPORT Point2D
{
public:
  inline Point2D();
  inline Point2D(double x, double y);
  inline Point2D(const QPoint& p);
  inline Point2D(const QPointF& p);
  Point2D(const VectorList<double>& p);

  double x() const {return _x[0];}
  double y() const {return _x[1];}
  const double * data() const {return _x;}
  double * data() {return _x;}
  double y(const CurvePointOptions *) const {return _x[1];}
  void setX(double v) {_x[0]=v;}
  void setY(double v) {_x[1]=v;}
  void setY(double v, const CurvePointOptions *) {_x[1]=v;}
  double at(int index) const;
  void setValid(bool) {}
  bool isValid() const {return true;}
  void average(const Point2D& p) {_x[1]=0.5*(_x[1]+p._x[1]);}

  void degreesToDMS();
  void DMSToDegrees();
  void degreesToDM();
  void DMToDegrees();
  void geographicalToRectangular(const Point2D& ref, double rotation=0.0);
  void rectangularToGeographical(const Point2D& ref, double rotation=0.0);
  void geographicalToUtm(UtmZone& zone);
  void utmToGeographical(const UtmZone& zone);

  inline void set(double xi, double yi);
  inline QPointF pointF() const {return QPointF(_x[0], _x[1]);}

  inline bool operator<(const Point2D& p) const;
  inline bool operator>(const Point2D& p) const;
  inline bool operator==(const Point2D& p) const;
  inline bool operator!=(const Point2D& obj) const;
  inline short compare(const Point2D& p) const;
  inline bool isSimilar(const Point2D& p, double tolerance) const;

  bool fromString(const StringSection& str);
  QString toString(char format='g', int precision=6) const;

  inline void operator+=(const Point2D& p);
  inline void operator-=(const Point2D& p);
  inline void operator*=(const Point2D& p);
  inline void operator*=(double mul);
  inline void operator/=(const Point2D& p);
  void operator/=(double div) {operator*=(1.0/div);}
  void operator*=(const Matrix2x2& transformation);
  inline double scalarProduct(const Point2D& p) const;
  inline double scalarProduct(const double * p) const;
  inline void interpole(double valX, const Point2D& p1, const Point2D& p2,
                        SamplingOptions xSampling, SamplingOptions ySampling);
  void translate(const Point2D& p) {operator+=(p);}
  void rotate(const Angle& a);

  inline Point2D operator+(const Point2D& p) const;
  inline Point2D operator-(const Point2D& p) const;
  inline Point2D operator-() const;
  inline Point2D operator*(double mul) const;
  inline Point2D operator*(const Point2D& p) const;
  inline Point2D operator/(double div) const {return operator*(1.0/div);}
  inline Point2D operator/(const Point2D& p) const;

  bool isHit(const Point2D &p,double tolX, double tolY) const;
  double length() const {return ::sqrt(abs2());}
  double distanceTo(const Point2D &p) const;
  double squareDistanceTo(const Point2D &p) const;
  double distanceToSegment(const Point2D &p1, const Point2D &p2) const;
  double azimuth() const;
  double azimuthTo(const Point2D &p) const;
  double geographicalDistanceTo(const Point2D &p) const;
  double geographicalAzimuthTo(const Point2D &p) const;
  void move(double distance, const Angle& azimuth);
  void round(double maximum=1.0);
  inline double abs2() const;
  inline void abs();
  inline void yLog();
  inline void yExp();
  inline void yInverse();
private:
  static double utmM(double phi);

  double _x[2];
};

QGPCOREMATH_EXPORT uint qHash(Point2D key);

} // namespace QGpCoreMath

namespace QGpCoreMath {

inline Point2D::Point2D()
{
  _x[0]=0.;
  _x[1]=0.;
}

inline Point2D::Point2D(double xi, double yi)
{
  _x[0]=xi;
  _x[1]=yi;
}

inline Point2D::Point2D(const QPoint& p)
{
  _x[0]=p.x();
  _x[1]=p.y();
}

inline Point2D::Point2D(const QPointF& p)
{
  _x[0]=p.x();
  _x[1]=p.y();
}

inline void Point2D::set(double xi, double yi)
{
  _x[0]=xi;
  _x[1]=yi;
}

inline bool Point2D::operator<(const Point2D& p) const
{
  if(_x[0]<p._x[0]) {
    return true;
  } else if(_x[0]==p._x[0] && _x[1]<p._x[1]) {
    return true;
  }
  return false;
}

inline bool Point2D::operator>(const Point2D& p) const
{
  if(_x[0]>p._x[0]) {
    return true;
  } else if(_x[0]==p._x[0] && _x[1]>p._x[1]) {
    return true;
  }
  return false;
}

inline bool Point2D::operator==(const Point2D& p) const
{
  return _x[0]==p._x[0] && _x[1]==p._x[1];;
}

inline bool Point2D::operator!=(const Point2D& p) const
{
  return _x[0]!=p._x[0] || _x[1]!=p._x[1];;
}

inline short Point2D::compare (const Point2D& p) const
{
  if(_x[0]<p._x[0]) return(-1);
  else if(_x[0]>p._x[0]) return(1);
  else if(_x[1]<p._x[1]) return(-1);
  else if(_x[1]>p._x[1]) return(1);
  else return(0);
}

inline Point2D Point2D::operator+(const Point2D& p) const
{
  return Point2D(_x[0]+p._x[0], _x[1]+p._x[1]);
}

inline Point2D Point2D::operator-(const Point2D& p) const
{
  return Point2D(_x[0]-p._x[0], _x[1]-p._x[1]);
}

inline Point2D Point2D::operator-() const
{
  return Point2D(-_x[0], -_x[1]);
}

inline Point2D Point2D::operator*(double mul) const
{
  return Point2D(_x[0]*mul, _x[1]*mul);
}

inline Point2D Point2D::operator*(const Point2D& p) const
{
  return Point2D(_x[0]*p._x[0], _x[1]*p._x[1]);
}

inline Point2D Point2D::operator/(const Point2D& p) const
{
  return Point2D(_x[0]/p._x[0], _x[1]/p._x[1]);
}

inline void Point2D::operator+=(const Point2D& p)
{
  _x[0]+=p._x[0];
  _x[1]+=p._x[1];
}

inline void Point2D::operator-=(const Point2D& p)
{
  _x[0]-=p._x[0];
  _x[1]-=p._x[1];
}

inline void Point2D::operator*=(const Point2D& p)
{
  _x[0]*=p._x[0];
  _x[1]*=p._x[1];
}

inline void Point2D::operator*=(double mul)
{
  _x[0]*=mul;
  _x[1]*=mul;
}

inline void Point2D::operator/=(const Point2D& p)
{
  _x[0]/=p._x[0];
  _x[1]/=p._x[1];
}

inline double Point2D::scalarProduct(const Point2D& p) const
{
  return _x[0]*p._x[0]+_x[1]*p._x[1];
}

inline double Point2D::scalarProduct(const double * p) const
{
  return _x[0]*p[0]+_x[1]*p[1];
}

inline void Point2D::interpole(double valX, const Point2D& p1, const Point2D& p2,
                               SamplingOptions xSampling, SamplingOptions ySampling)
{
  setX(valX);
  double x1, x2;
  double y1, y2;
  if(xSampling & InverseScale) {
    valX=1.0/valX;
    x1=1.0/p1.x();
    x2=1.0/p2.x();
  } else if(xSampling & LogScale) {
    valX=log(valX);
    x1=log(p1.x());
    x2=log(p2.x());
  } else {
    x1=p1.x();
    x2=p2.x();
  }
  double factor=(valX-x1)/(x2-x1);
  if(ySampling & InverseScale) {
    y1=1.0/p1.y();
    y2=1.0/p2.y();
    this->setY(1.0/(y1+(y2-y1)*factor));
  } else if(ySampling & LogScale) {
    y1=log(p1.y());
    y2=log(p2.y());
    this->setY(exp(y1+(y2-y1)*factor));
  } else {
    y1=p1.y();
    y2=p2.y();
    this->setY(y1+(y2-y1)*factor);
  }
}

inline bool Point2D::isSimilar(const Point2D& p, double tolerance) const
{
  return fabs(p.x()-x())<tolerance &&
         fabs(p.y()-y())<tolerance;
}

inline void Point2D::abs()
{
  _x[0]=fabs(_x[0]);
  _x[1]=fabs(_x[1]);
}

inline double Point2D::abs2() const
{
  return _x[0]*_x[0]+_x[1]*_x[1];
}

inline void Point2D::yLog()
{
  _x[1]=::log(_x[1]);
}

inline void Point2D::yExp()
{
  _x[1]=::exp(_x[1]);
}

inline void Point2D::yInverse()
{
  _x[1]=1.0/_x[1];
}

QGPCOREMATH_EXPORT QTextStream& operator<< (QTextStream& s, const QList<Point2D>& plist);
QGPCOREMATH_EXPORT QTextStream& operator>> (QTextStream& s, QList<Point2D>& plist);
QGPCOREMATH_EXPORT QTextStream& operator<< (QTextStream& s, const Point2D& p);
QGPCOREMATH_EXPORT QTextStream& operator>> (QTextStream& s, Point2D& p);
QGPCOREMATH_EXPORT QDataStream& operator<< (QDataStream& s, const Point2D& p);
QGPCOREMATH_EXPORT QDataStream& operator>> (QDataStream& s, Point2D& p);

} // namespace QGpCoreMath

#endif // POINT2D.H
